@@ -227,9 +227,12 @@
     Updated by      Richard Gooch   25-NOV-2001: Created <<event_types>> table.
   Generate synthetic REGISTER event for directories in <do_scan_and_service>.
 
-    Last updated by Richard Gooch   28-NOV-2001: Renamed <make_symlink> to
+    Updated by      Richard Gooch   28-NOV-2001: Renamed <make_symlink> to
   <mksymlink> and made it global. Changed to return 0 on "success".
 
+    Last updated by Richard Gooch   29-NOV-2001: Changed <action_copy> to set
+  sticky bit as required. Added "RESTORE" key.
+
 
 */
 #define _GNU_SOURCE
@@ -290,7 +293,8 @@
 #define AC_MKOLDCOMPAT              7
 #define AC_MKNEWCOMPAT              8
 #define AC_RMOLDCOMPAT              9
-#define AC_RMNEWCOMPAT              10
+#define AC_RMNEWCOMPAT             10
+#define AC_RESTORE                 11
 
 
 struct permissions_type
@@ -408,6 +412,10 @@ static void action_copy (const struct devfsd_notify_struct *info,
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action);
 static void free_config ();
+static void do_restore (const char *dirname, int rootlen);
+static flag copy_inode (const char *destpath, const struct stat *deststat,
+			mode_t new_mode,
+			const char *sourcepath, const struct stat *sourcestat);
 static void do_debug (int fd);
 static uid_t get_uid (const char *string);
 static gid_t get_gid (const char *string);
@@ -796,6 +804,11 @@ static void process_config_line (CONST char *line, unsigned long *event_mask)
 		     event_mask);
 	return;
     }
+    if (strcasecmp (when, "RESTORE") == 0)
+    {
+	do_restore ( name, strlen (name) );
+	return;
+    }
     if (num_args < 3)
     {
 	SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
@@ -1347,10 +1360,12 @@ static void action_copy (const struct devfsd_notify_struct *info,
     [RETURNS] Nothing.
 */
 {
+    mode_t new_mode;
     struct get_variable_info gv_info;
     struct stat source_stat, dest_stat;
     char source[STRING_LENGTH], destination[STRING_LENGTH];
 
+    if ( (info->type == DEVFSD_NOTIFY_CHANGE) && S_ISLNK (info->mode) ) return;
     gv_info.info = info;
     gv_info.devname = info->devname;
     snprintf (gv_info.devpath, sizeof (gv_info.devpath), "%s/%s",
@@ -1371,85 +1386,13 @@ static void action_copy (const struct devfsd_notify_struct *info,
     }
     if (lstat (source, &source_stat) != 0) return;
     if (lstat (destination, &dest_stat) != 0) dest_stat.st_mode = 0;
-    if ( (source_stat.st_mode & S_IFMT) == (dest_stat.st_mode & S_IFMT) )
-    {
-	/*  Same type  */
-	if ( S_ISLNK (source_stat.st_mode) )
-	{
-	    int source_len, dest_len;
-	    char source_link[STRING_LENGTH], dest_link[STRING_LENGTH];
-
-	    if ( ( source_len = readlink (source, source_link,
-					  STRING_LENGTH - 1) ) < 0 ) return;
-	    source_link[source_len] = '\0';
-	    if ( ( dest_len = readlink (destination, dest_link,
-					STRING_LENGTH - 1) ) < 0 ) return;
-	    dest_link[dest_len] = '\0';
-	    if ( (source_len != dest_len) ||
-		 (strcmp (source_link, dest_link) != 0) )
-	    {
-		unlink (destination);
-		symlink (source_link, destination);
-	    }
-	    return;
-	}   /*  Else not a symlink  */
-	if (source_stat.st_mode != dest_stat.st_mode)
-	    chmod (destination, source_stat.st_mode & ~S_IFMT);
-	if ( (source_stat.st_uid != dest_stat.st_uid) ||
-	     (source_stat.st_gid != dest_stat.st_gid) )
-	    chown (destination, source_stat.st_uid, source_stat.st_gid);
-	return;
-    }
-    /*  Different types: unlink and create  */
-    unlink (destination);
-    switch (source_stat.st_mode & S_IFMT)
-    {
-	int fd, val;
-	struct sockaddr_un un_addr;
-	char symlink_val[STRING_LENGTH];
-
-      case S_IFSOCK:
-	if ( ( fd = socket (AF_UNIX, SOCK_STREAM, 0) ) < 0 ) break;
-	un_addr.sun_family = AF_UNIX;
-	snprintf (un_addr.sun_path, sizeof (un_addr.sun_path), "%s",
-		  destination);
-	val = bind (fd, (struct sockaddr *) &un_addr, (int) sizeof un_addr);
-	close (fd);
-	if (val != 0) break;
-	if (chmod (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
-	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
-	    return;
-	break;
-      case S_IFLNK:
-	if ( ( val = readlink (source, symlink_val, STRING_LENGTH - 1) ) < 0 )
-	    break;
-	symlink_val[val] = '\0';
-	if (symlink (symlink_val, destination) == 0) return;
-	break;
-      case S_IFREG:
-	if ( ( fd = open (destination, O_RDONLY | O_CREAT,
-			  source_stat.st_mode & ~S_IFMT) ) < 0 ) break;
-	close (fd);
-	if (chmod (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
-	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
-	    return;
-	break;
-      case S_IFBLK:
-      case S_IFCHR:
-      case S_IFIFO:
-	if (mknod (destination, source_stat.st_mode, source_stat.st_rdev) != 0)
-	    break;
-	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
-	    return;
-	break;
-      case S_IFDIR:
-	if (mkdir (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
-	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
-	    return;
-	return;
-	/*break;*/
-    }
-    SYSLOG (LOG_ERR, "error copying: \"%s\" to \"%s\"\n", source, destination);
+    new_mode = source_stat.st_mode & ~S_ISVTX;
+    if (info->type == DEVFSD_NOTIFY_CREATE) new_mode |= S_ISVTX;
+    else if ( (info->type == DEVFSD_NOTIFY_CHANGE) &&
+	      (dest_stat.st_mode & S_ISVTX) ) new_mode |= S_ISVTX;
+    if ( !copy_inode (destination, &dest_stat, new_mode, source,&source_stat) )
+	SYSLOG (LOG_ERR, "error copying: \"%s\" to \"%s\"\n",
+		source, destination);
 }   /*  End Function action_copy  */
 
 static void action_compat (const struct devfsd_notify_struct *info,
@@ -1576,6 +1519,141 @@ static void action_compat (const struct devfsd_notify_struct *info,
     }
 }   /*  End Function action_compat  */
 
+static void do_restore (const char *dirname, int rootlen)
+/*  [SUMMARY] Restore state from a directory.
+    <dirname> The directory containing the backing store.
+    <rootlen> The length of the root of the state directory hierarchy.
+    [RETURNS] Nothing.
+*/
+{
+    DIR *dp;
+    struct dirent *de;
+
+    if ( ( dp = opendir (dirname) ) == NULL )
+    {
+	SYSLOG (LOG_ERR, "Error opening directory \"%s\"\t%s\n",
+		dirname, ERRSTRING);
+	return;
+    }
+    while ( (de = readdir (dp) ) != NULL )
+    {
+	mode_t new_mode;
+	struct stat source_stat, dest_stat;
+	char spath[STRING_LENGTH], dpath[STRING_LENGTH];
+
+	if ( (strcmp (de->d_name, ".") == 0) ||
+	     (strcmp (de->d_name, "..") == 0) ) continue;
+	snprintf (spath, sizeof spath, "%s/%s", dirname, de->d_name);
+	if (lstat (spath, &source_stat) != 0)
+	{
+	    SYSLOG (LOG_ERR, "Error lstat(2)ing file \"%s\"\t%s\n",
+		    spath, ERRSTRING);
+	    continue;
+	}
+	snprintf (dpath, sizeof dpath, "%s%s", mount_point, spath + rootlen);
+	if (lstat (dpath, &dest_stat) != 0) dest_stat.st_mode = 0;
+	new_mode = source_stat.st_mode & ~S_ISVTX;
+	if ( S_ISLNK (source_stat.st_mode) || (source_stat.st_mode & S_ISVTX) )
+	    copy_inode (dpath, &dest_stat, new_mode, spath, &source_stat);
+	if ( S_ISDIR (source_stat.st_mode) ) do_restore (spath, rootlen);
+    }
+    closedir (dp);
+}   /*  End Function do_restore  */
+
+static flag copy_inode (const char *destpath, const struct stat *dest_stat,
+			mode_t new_mode,
+			const char *sourcepath, const struct stat *source_stat)
+/*  [SUMMARY] Copy an inode.
+    <destpath> The destination path. An existing inode may be deleted.
+    <dest_stat> The destination stat(2) information.
+    <new_mode> The desired new mode for the destination.
+    <sourcepath> The source path.
+    <source_stat> The source stat(2) information.
+    [RETURNS] TRUE on success, else FALSE.
+*/
+{
+    if ( (source_stat->st_mode & S_IFMT) == (dest_stat->st_mode & S_IFMT) )
+    {
+	/*  Same type  */
+	if ( S_ISLNK (source_stat->st_mode) )
+	{
+	    int source_len, dest_len;
+	    char source_link[STRING_LENGTH], dest_link[STRING_LENGTH];
+
+	    if ( ( source_len = readlink (sourcepath, source_link,
+					  STRING_LENGTH - 1) ) < 0 )
+		return (FALSE);
+	    source_link[source_len] = '\0';
+	    if ( ( dest_len = readlink (destpath, dest_link,
+					STRING_LENGTH - 1) ) < 0 )
+		return (FALSE);
+	    dest_link[dest_len] = '\0';
+	    if ( (source_len != dest_len) ||
+		 (strcmp (source_link, dest_link) != 0) )
+	    {
+		unlink (destpath);
+		symlink (source_link, destpath);
+	    }
+	    return (TRUE);
+	}   /*  Else not a symlink  */
+	if (new_mode != dest_stat->st_mode)
+	    chmod (destpath, new_mode & ~S_IFMT);
+	if ( (source_stat->st_uid != dest_stat->st_uid) ||
+	     (source_stat->st_gid != dest_stat->st_gid) )
+	    chown (destpath, source_stat->st_uid, source_stat->st_gid);
+	return (TRUE);
+    }
+    /*  Different types: unlink and create  */
+    unlink (destpath);
+    switch (source_stat->st_mode & S_IFMT)
+    {
+	int fd, val;
+	struct sockaddr_un un_addr;
+	char symlink_val[STRING_LENGTH];
+
+      case S_IFSOCK:
+	if ( ( fd = socket (AF_UNIX, SOCK_STREAM, 0) ) < 0 ) break;
+	un_addr.sun_family = AF_UNIX;
+	snprintf (un_addr.sun_path, sizeof (un_addr.sun_path), "%s",
+		  destpath);
+	val = bind (fd, (struct sockaddr *) &un_addr, (int) sizeof un_addr);
+	close (fd);
+	if (val != 0) break;
+	if (chmod (destpath, new_mode & ~S_IFMT) != 0) break;
+	if (chown (destpath, source_stat->st_uid, source_stat->st_gid) == 0)
+	    return (TRUE);
+	break;
+      case S_IFLNK:
+	if ( ( val = readlink (sourcepath, symlink_val, STRING_LENGTH - 1) )
+	     < 0 ) break;
+	symlink_val[val] = '\0';
+	if (symlink (symlink_val, destpath) == 0) return (TRUE);
+	break;
+      case S_IFREG:
+	if ( ( fd = open (destpath, O_RDONLY | O_CREAT,
+			  new_mode & ~S_IFMT) ) < 0 ) break;
+	close (fd);
+	if (chmod (destpath, new_mode & ~S_IFMT) != 0) break;
+	if (chown (destpath, source_stat->st_uid, source_stat->st_gid) == 0)
+	    return (TRUE);
+	break;
+      case S_IFBLK:
+      case S_IFCHR:
+      case S_IFIFO:
+	if (mknod (destpath, new_mode, source_stat->st_rdev) != 0) break;
+	if (chown (destpath, source_stat->st_uid, source_stat->st_gid) == 0)
+	    return (TRUE);
+	break;
+      case S_IFDIR:
+	if (mkdir (destpath, new_mode & ~S_IFMT) != 0) break;
+	if (chown (destpath, source_stat->st_uid, source_stat->st_gid) == 0)
+	    return (TRUE);
+	return (TRUE);
+	/*break;*/
+    }
+    return (FALSE);
+}   /*  End Function copy_inode  */
+
 static void free_config ()
 /*  [SUMMARY] Free the configuration information.
     [RETURNS] Nothing.
